pub static STOP_WORDS: &[&str] = &[
    "a", "an", "the", "and", "or", "of", "in", "to", "by", "at", "on", "with", "is", "are", "was",
    "were", "be", "been", "being", "it", "its", "as", "all", "each", "into", "up", "out", "do",
    "does", "did", "have", "has", "had", "they", "them", "their", "there", "this", "these",
    "those", "that",
];
